package extensions;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.net.URLDecoder;
import java.net.Socket;
import java.net.URLEncoder;
import java.net.URLDecoder;
import java.text.DateFormat;
import java.util.Calendar;
import java.util.Collections;
import java.util.Map;
import java.util.HashMap;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.Main;
import dareka.common.CloseUtil;
import dareka.common.DiskFreeSpace;
import dareka.common.Logger;
import dareka.common.Pair;
import dareka.extensions.Extension2;
import dareka.extensions.ExtensionManager;
import dareka.extensions.RequestFilter;
import dareka.extensions.Rewriter;
import dareka.processor.HttpRequestHeader;
import dareka.processor.HttpResponseHeader;
import dareka.processor.Resource;
import dareka.processor.URLResource;
import dareka.processor.StringResource;
import dareka.processor.TransferListener;
import dareka.processor.Processor;
import dareka.processor.impl.Cache;
import dareka.processor.impl.NicoCachingTitleRetriever;
import dareka.processor.util.GetThumbInfoUtil;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;

public class UnlimitedMyList implements Extension2, Processor {
	public static final String LOG_PREFIX = "ml";
	public static final String PROP_DEBUG = "myListDebug";
	private static final String[] PROCESSOR_SUPPORTED_METHODS = new String[] { "GET" };
	private static final Pattern PROCESSOR_SUPPORTED_PATTERN = Pattern.compile(
			"^(http://[^/]+\\.(?:nico|smile)video\\.jp)/umylist(/.*)$");
	private static final Pattern PATH_SMILE_PATTERN = Pattern.compile(
			"^/smile\\?s=(\\d+)\\.\\d+$");
	private static final Pattern PATH_SWFHEADER_PATTERN = Pattern.compile(
			"^/(?:madd:add)\\?(?:nm)?(\\d+)$");
	private static ExtUtil util;
	private static Connection con;
	private Statement stmt;

	public void registerExtensions(ExtensionManager mgr) {
		if( util == null ) {
			util = new ExtUtil(this, LOG_PREFIX, PROP_DEBUG);
		}
		/* クラスの読み込みよくわからないので誤魔化した */
		try {
			URL[] urls = new URL[1];
			
			urls[0] = new URL("jar:file:/sqlite-jdbc.jar!/");
			
			ClassLoader url = new URLClassLoader(urls);
			Class.forName("org.sqlite.JDBC");
		} catch(Exception e) {
		}
		if( con == null ) {
			try {
				con = DriverManager.getConnection("jdbc:sqlite:umylist.db");	
				stmt = con.createStatement();
				ResultSet rs;
				rs = stmt.executeQuery("SELECT count(*) FROM sqlite_master WHERE type='table' AND name='umylist';");
				rs.next();
				// テーブルなかったら作る
				if( rs.getInt(1) == 0 ) {
					stmt.execute("create table umylist(id INTEGER PRIMARY KEY AUTOINCREMENT,name text)");
				}
				rs = stmt.executeQuery("SELECT count(*) FROM sqlite_master WHERE type='table' AND name='umovie';");
				rs.next();
				if( rs.getInt(1) == 0 ) {
					stmt.execute("create table umovie(id INTEGER PRIMARY KEY AUTOINCREMENT,mylistid integer,movieid text)");
				}
			} catch(Exception e) {
				util.info(e+"");
			}
		}
		mgr.registerProcessor(this);      // Processorとして登録
	}
	
	/**
	 * @return バージョン文字列。長いのは自重するべき。
	 */
	public String getVersionString() {
		return "Unlimited MyList 1.0";
	}

	public String[] getSupportedMethods() {
		return PROCESSOR_SUPPORTED_METHODS;
	}

	public Pattern getSupportedURLAsPattern() {
		return PROCESSOR_SUPPORTED_PATTERN;
	}
	
	public String getSupportedURLAsString() {
		return null;
	}
	
	/*
	*/
	private HashMap getQueryParameter(String s) {
		HashMap<String,String> map = new HashMap<String,String>();
		String[] v = s.split("&");
		for(int i = 0; i < v.length; i++) {
			String[] w = v[i].split("=");
			if( w.length == 2 ) {
				map.put(w[0],w[1]);
			} else {
				map.put(w[0],"");
			}
		}
		return map;
	}
	
	private String replace(String str) {
		String ret = new String(str);
		String[] escape = {"&", "<", ">", "\"", "\'"};
		String[] replace = {"&amp;", "&lt;", "&gt;", "&quot;", "&#39;"};
		for ( int i=0; i < escape.length; i++ ){
			ret = ret.replace(escape[i], replace[i]);
		}
		return ret;
	}
	
	private boolean isAccept(String id) {
		Pattern pat = Pattern.compile("^(sm|so|nm)\\d+$");
		Matcher mt = pat.matcher(id);
		return mt.matches();
	}

	public Resource onRequest(HttpRequestHeader requestHeader, Socket browser)
	throws IOException {
		try {
			Matcher m = PROCESSOR_SUPPORTED_PATTERN.matcher(requestHeader.getURI());
			if (!m.matches()) {
				throw new IOException("");
			}
			if( m.groupCount() < 2 ) {
				return new StringResource("err");
			}
			String a = m.group(1), b = m.group(2);
			String[] tmp = b.split("\\?");
			String path = tmp[0];
			HashMap get = new HashMap();
			if( tmp.length != 1 ) {
				get = getQueryParameter(tmp[1]);
			}
			String id = "";
			if( path.equals("/list") ) {
				id += "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n";
				if( get.get("v") == null ) {
					id += "<mylist>\n";
					ResultSet rs = stmt.executeQuery("select * from umylist");
					while( rs.next() ) {
						id += "\t<list>\n";
						id += "\t\t<id>"+rs.getString("id")+"</id>\n";
						id += "\t\t<name>"+rs.getString("name")+"</name>\n";
						id += "\t</list>\n";
					}
					rs.close();
					id += "</mylist>";
				} else {
					String mid = (String)get.get("v");
					id += "<mylist>\n";
					String sql = "select * from umovie where mylistid=?";
					PreparedStatement pstmt = con.prepareStatement(sql);
					pstmt.setString(1,replace(mid));
					ResultSet rs = pstmt.executeQuery();
					while( rs.next() ) {
						id += "\t<list>\n";
						id += "\t\t<id>"+rs.getString("movieid")+"</id>\n";
						id += "\t</list>\n";
					}
					rs.close();
					id += "</mylist>\n";
				}
				return new StringResource(id);
			} else if( path.equals("/madd") ) {
				if( get.get("name") == null ) {
					return new StringResource("faild");
				}
				String sql = "INSERT INTO umylist (name) VALUES(?)";
				PreparedStatement pstmt = con.prepareStatement(sql);
				String name = (String)URLDecoder.decode((String)get.get("name"),"utf8");
				pstmt.setString(1,replace(name));
				pstmt.execute();
				return new StringResource("succeeded");
			} else if( path.equals("/add") ) {
				if( get.get("mid") != null && get.get("mylist") != null ) {

					String mylist = replace((String)get.get("mylist"));
					String mid = replace((String)get.get("mid"));
					if( !isAccept(mid) ) {
						return new StringResource("invalidID");
					}
					String sql = "SELECT count(*) FROM umovie WHERE mylistid=? AND movieid=?";
					PreparedStatement pstmt = con.prepareStatement(sql);
					pstmt.setString(1,mylist);
					pstmt.setString(2,mid);
					ResultSet rs;
					rs = pstmt.executeQuery();
					rs.next();
					// 動画がすでに入っている
					if( rs.getInt(1) != 0 ) {
						return new StringResource("overlap");
					}
					sql = "SELECT count(*) FROM umylist WHERE id=?";
					pstmt = con.prepareStatement(sql);
					pstmt.setString(1,mylist);
					rs = pstmt.executeQuery();
					rs.next();
					// 存在しないマイリスト
					if( rs.getInt(1) != 1 ) {
						return new StringResource("invalidMylist");
					}

					sql = "INSERT INTO umovie (mylistid,movieid) VALUES(?,?)";
					pstmt = con.prepareStatement(sql);
					pstmt.setString(1,mylist);
					pstmt.setString(2,mid);
					pstmt.execute();
					return new StringResource("succeeded");
				} else {
					return new StringResource("faild");
				}
			} else if( path.equals("/remove") ) {
				if( get.get("mid") != null && get.get("mylist") != null ) {
					String mylist = replace((String)get.get("mylist"));
					String mid = replace((String)get.get("mid"));
					String sql = "DELETE FROM umovie WHERE mylistid=? AND movieid=?";
					PreparedStatement pstmt = con.prepareStatement(sql);
					pstmt.setString(1,mylist);
					pstmt.setString(2,mid);
					pstmt.execute();
					return new StringResource("succeeded");
				} else {
					return new StringResource("faild");
				}
			} else if( path.equals("/mremove") ) {
				if( get.get("mylist") != null ) {
					String mylist = replace((String)get.get("mylist"));
					String sql = "DELETE FROM umovie WHERE mylistid=?";
					PreparedStatement pstmt = con.prepareStatement(sql);
					pstmt.setString(1,mylist);
					pstmt.execute();
					sql = "DELETE FROM umylist WHERE id=?";
					pstmt = con.prepareStatement(sql);
					pstmt.setString(1,mylist);
					pstmt.execute();
					return new StringResource("succeeded");
				} else {
					return new StringResource("faild");
				}
			}
		} catch(Exception e) {
			return new StringResource(e+"");
		}
		return new StringResource("");
	}
}